#!/bin/sh
src=`pwd`
tmp=$src/tests-tmp
rm -rf $tmp
mkdir -p $tmp
PATH="$src:/bin:/usr/bin:/usr/local/bin"
tests_failed=0
tests_count=0
sfqmail() { smtpfront-qmail 2>/dev/null | tail +2; }
sfecho() { smtpfront-echo 2>/dev/null | tail +2; }
pfauth() { pop3front-auth "$@" echo Yes. 2>/dev/null | tail +2; }
pfmaildir() { pop3front-maildir "$@" 2>/dev/null | tail +2; }
maildir=$tmp/Maildir
maildir() {
  rm -rf $maildir
  mkdir -p $maildir/{cur,new,tmp}
}
tstmsg() {
  fn=$1
  {
    echo "Header: foo"
    echo
    echo "body"
  } >$maildir/$fn
}

run_compare_test() {
  local name=$1
  shift
  sed -e "s:@SOURCE@:$src:g" >$tmp/expected
  ( runtest "$@" ) 2>&1 >$tmp/actual-raw
  cat -v $tmp/actual-raw >$tmp/actual
  if ! cmp $tmp/expected $tmp/actual >/dev/null 2>&1
  then
    echo "Test $name $@ failed:"
	( cd $tmp; diff -u9999 expected actual )
	let tests_failed=tests_failed+1
  fi
  rm -f $tmp/expected $tmp/actual
  let tests_count=tests_count+1
}

##### Test tests/smtpfront-auth-plain #####

runtest() {
sfqmail <<EOF
AUTH PLAIN dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
EOF

export CVM_SASL_PLAIN=$src/cvm/cvm-unix

sfqmail <<EOF
MAIL FROM: <user@remote>
RCPT TO: <user@remote>
AUTH PLAIN dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3x=
AUTH PLAIN dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
AUTH PLAIN dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
MAIL FROM: <user@remote>
RCPT TO: <user@remote>
EOF

sfqmail << EOF
AUTH PLAIN
dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
EOF

sfqmail <<EOF
AUTH PLAIN
*
MAIL FROM: <user@remote>
RCPT TO: <user@remote>
EOF

sfqmail << EOF
AUTH PLAIN XXXXdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
EOF

unset CVM_SASL_PLAIN
}
run_compare_test tests/smtpfront-auth-plain  <<END_OF_TEST_RESULTS
504 Unrecognized authentication mechanism.^M
250 Sender accepted.^M
553 Sorry, that domain isn't in my list of allowed rcpthosts.^M
501 Authentication failed.^M
235 Authentication succeeded.^M
503 You are already authenticated.^M
250 Sender accepted.^M
250 Recipient accepted.^M
334 ^M
235 Authentication succeeded.^M
334 ^M
501 Authentication failed.^M
250 Sender accepted.^M
553 Sorry, that domain isn't in my list of allowed rcpthosts.^M
235 Authentication succeeded.^M
END_OF_TEST_RESULTS


##### Test tests/smtpfront-auth-login #####

runtest() {
sfqmail <<EOF
AUTH LOGIN
EOF

export CVM_SASL_LOGIN=$src/cvm/cvm-unix

sfqmail <<EOF
MAIL FROM: <user@remote>
RCPT TO: <user@remote>
AUTH LOGIN
dGVzdHVzZXI=
dGVzdHBhc3x=
AUTH LOGIN
dGVzdHVzZXI=
dGVzdHBhc3M=
AUTH LOGIN
MAIL FROM: <user@remote>
RCPT TO: <user@remote>
EOF

sfqmail << EOF
AUTH LOGIN dGVzdHVzZXI=
dGVzdHBhc3M=
EOF

sfqmail <<EOF
AUTH LOGIN
dGVzdHVzZXI=
*
MAIL FROM: <user@remote>
RCPT TO: <user@remote>
EOF

unset CVM_SASL_PLAIN
}
run_compare_test tests/smtpfront-auth-login  <<END_OF_TEST_RESULTS
504 Unrecognized authentication mechanism.^M
250 Sender accepted.^M
553 Sorry, that domain isn't in my list of allowed rcpthosts.^M
334 VXNlcm5hbWU6^M
334 UGFzc3dvcmQ6^M
501 Authentication failed.^M
334 VXNlcm5hbWU6^M
334 UGFzc3dvcmQ6^M
235 Authentication succeeded.^M
503 You are already authenticated.^M
250 Sender accepted.^M
250 Recipient accepted.^M
334 UGFzc3dvcmQ6^M
235 Authentication succeeded.^M
334 VXNlcm5hbWU6^M
334 UGFzc3dvcmQ6^M
501 Authentication failed.^M
250 Sender accepted.^M
553 Sorry, that domain isn't in my list of allowed rcpthosts.^M
END_OF_TEST_RESULTS


##### Test tests/pop3front-auth-userpass #####

runtest() {
# Should fix this and the others to actually check the stderr output too.
pfauth $src/cvm/cvm-unix <<EOF 2>/dev/null
USER testuser
PASS testpass
EOF
pfauth $src/cvm/cvm-unix <<EOF 2>/dev/null
USER testuser
PASS testpasx
EOF
}
run_compare_test tests/pop3front-auth-userpass  <<END_OF_TEST_RESULTS
+OK ^M
Yes.
+OK ^M
-ERR Authentication failed^M
END_OF_TEST_RESULTS


##### Test tests/pop3front-auth-plain #####

runtest() {
pfauth false <<EOF
AUTH PLAIN dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
EOF

export CVM_SASL_PLAIN=$src/cvm/cvm-unix

pfauth false <<EOF
AUTH PLAIN dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3x=
AUTH PLAIN dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
EOF

pfauth false <<EOF
AUTH PLAIN
dGVzdHVzZXIAdGVzdHVzZXIAdGVzdHBhc3M=
EOF

pfauth false <<EOF
AUTH PLAIN
*
EOF

unset CVM_SASL_PLAIN
}
run_compare_test tests/pop3front-auth-plain  <<END_OF_TEST_RESULTS
-ERR Unrecognized authentication mechanism.^M
-ERR Authentication failed.^M
Yes.
+ ^M
Yes.
+ ^M
-ERR Authentication failed.^M
END_OF_TEST_RESULTS


##### Test tests/pop3front-auth-login #####

runtest() {
pfauth false <<EOF
AUTH LOGIN
EOF

export CVM_SASL_LOGIN=$src/cvm/cvm-unix

pfauth false <<EOF
AUTH LOGIN
dGVzdHVzZXI=
dGVzdHBhc3x=
AUTH LOGIN
dGVzdHVzZXI=
dGVzdHBhc3M=
EOF

pfauth false <<EOF
AUTH LOGIN dGVzdHVzZXI=
dGVzdHBhc3M=
EOF

pfauth false <<EOF
AUTH LOGIN
dGVzdHVzZXI=
*
EOF

unset CVM_SASL_PLAIN
}
run_compare_test tests/pop3front-auth-login  <<END_OF_TEST_RESULTS
-ERR Unrecognized authentication mechanism.^M
+ VXNlcm5hbWU6^M
+ UGFzc3dvcmQ6^M
-ERR Authentication failed.^M
+ VXNlcm5hbWU6^M
+ UGFzc3dvcmQ6^M
Yes.
+ UGFzc3dvcmQ6^M
Yes.
+ VXNlcm5hbWU6^M
+ UGFzc3dvcmQ6^M
-ERR Authentication failed.^M
END_OF_TEST_RESULTS


##### Test tests/pop3front-auth #####

runtest() {
pop3front-auth false echo Yes. <<EOF 2>/dev/null
QUIT NO
QUIT
QUIT AGAIN
EOF
}
run_compare_test tests/pop3front-auth  <<END_OF_TEST_RESULTS
+OK ^M
-ERR Syntax error^M
+OK ^M
END_OF_TEST_RESULTS


##### Test tests/pop3front-maildir-state #####

runtest() {
local quit="$1"
local command="$2"
maildir
tstmsg new/1
(
  echo $command
  if $quit; then echo QUIT; fi
) | pfmaildir $maildir
( cd $maildir && find new cur -type f )
}
run_compare_test tests/pop3front-maildir-state 'false' 'UIDL' <<END_OF_TEST_RESULTS
+OK ^M
1 1^M
.^M
new/1
END_OF_TEST_RESULTS

run_compare_test tests/pop3front-maildir-state 'false' 'TOP 1 0' <<END_OF_TEST_RESULTS
+OK ^M
Header: foo^M
^M
^M
.^M
new/1
END_OF_TEST_RESULTS

run_compare_test tests/pop3front-maildir-state 'false' 'RETR 1' <<END_OF_TEST_RESULTS
+OK ^M
Header: foo^M
^M
body^M
^M
.^M
new/1
END_OF_TEST_RESULTS

run_compare_test tests/pop3front-maildir-state 'true' 'UIDL' <<END_OF_TEST_RESULTS
+OK ^M
1 1^M
.^M
+OK ^M
cur/1
END_OF_TEST_RESULTS

run_compare_test tests/pop3front-maildir-state 'true' 'TOP 1 0' <<END_OF_TEST_RESULTS
+OK ^M
Header: foo^M
^M
^M
.^M
+OK ^M
cur/1
END_OF_TEST_RESULTS

run_compare_test tests/pop3front-maildir-state 'true' 'RETR 1' <<END_OF_TEST_RESULTS
+OK ^M
Header: foo^M
^M
body^M
^M
.^M
+OK ^M
cur/1:2,S
END_OF_TEST_RESULTS


##### Test tests/pop3front-maildir-flags #####

runtest() {
pftest() {
  (
    for line in "$@"; do
      echo $line
    done
  ) | pfmaildir $maildir
  ( cd $maildir && ls -1 */* )
}

# Does it properly parse existing v2 flags
maildir
tstmsg new/1000000000.12345.here:2,FX
pftest 'RETR 1' QUIT

# Does it properly ignore non-v2 flags
maildir
tstmsg new/1000000000.12345.here:1fsd
pftest 'RETR 1' QUIT
}
run_compare_test tests/pop3front-maildir-flags  <<END_OF_TEST_RESULTS
+OK ^M
Header: foo^M
^M
body^M
^M
.^M
+OK ^M
cur/1000000000.12345.here:2,FXS
+OK ^M
Header: foo^M
^M
body^M
^M
.^M
+OK ^M
cur/1000000000.12345.here:1fsd
END_OF_TEST_RESULTS


##### Test tests/pop3front-maildir-sort #####

runtest() {
maildir
tstmsg new/99
tstmsg cur/100
tstmsg new/101
tstmsg cur/98

echo UIDL | pfmaildir $maildir
}
run_compare_test tests/pop3front-maildir-sort  <<END_OF_TEST_RESULTS
+OK ^M
1 98^M
2 99^M
3 100^M
4 101^M
.^M
END_OF_TEST_RESULTS


##### Test tests/pop3front-maildir-last #####

runtest() {
pflast() { echo LAST | pfmaildir $maildir; }
maildir
tstmsg new/1
tstmsg new/2
pflast
mv $maildir/new/1 $maildir/cur/1
pflast
mv $maildir/cur/1 $maildir/cur/1:2,S
pflast
mv $maildir/new/2 $maildir/cur/2:2,S
pflast
mv $maildir/cur/1:2,S $maildir/new/1
pflast
}
run_compare_test tests/pop3front-maildir-last  <<END_OF_TEST_RESULTS
+OK 0^M
+OK 0^M
+OK 1^M
+OK 2^M
+OK 2^M
END_OF_TEST_RESULTS


##### Test tests/smtpgreeting #####

runtest() {
env SMTPGREETING='hostname hello' smtpfront-echo 2>/dev/null </dev/null
}
run_compare_test tests/smtpgreeting  <<END_OF_TEST_RESULTS
220 hostname hello ESMTP^M
END_OF_TEST_RESULTS


##### Test tests/smtpfront-bad-bounce #####

runtest() {
sfecho <<EOF
MAIL FROM:<not@bounce>
RCPT TO:<addr1@domain>
RCPT TO:<addr2@domain>
DATA
.
EOF

sfecho <<EOF
MAIL FROM:<>
RCPT TO:<addr1@domain>
DATA
.
EOF

sfecho <<EOF
MAIL FROM:<>
RCPT TO:<addr1@domain>
RCPT TO:<addr2@domain>
DATA
EOF
}
run_compare_test tests/smtpfront-bad-bounce  <<END_OF_TEST_RESULTS
250 Sender='not@bounce'.^M
250 Recipient='addr1@domain'.^M
250 Recipient='addr2@domain'.^M
354 End your message with a period.^M
250 Received 94 bytes.^M
250 Sender=''.^M
250 Recipient='addr1@domain'.^M
354 End your message with a period.^M
250 Received 94 bytes.^M
250 Sender=''.^M
250 Recipient='addr1@domain'.^M
550 Bounce messages should have a single recipient.^M
550 Bounce messages should have a single recipient.^M
END_OF_TEST_RESULTS


##### Test tests/rules-sender #####

runtest() {
cat >$tmp/rules <<EOF
kone@one:*:KK
dtwo@two:*:DD
zthree@three:*:ZZ
zfour@four:one@one:ZZ
pfive@five:*:PP
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<one@one>
MAIL FROM:<two@two>
MAIL FROM:<three@three>
MAIL FROM:<four@four>
MAIL FROM:<five@five>
EOF

rm -f $tmp/rules
}
run_compare_test tests/rules-sender  <<END_OF_TEST_RESULTS
250 KK^M
553 DD^M
451 ZZ^M
250 Sender='four@four'.^M
250 Sender='five@five'.^M
END_OF_TEST_RESULTS


##### Test tests/rules-recip #####

runtest() {
cat >$tmp/rules <<EOF
k*:one@one:KK
d*:two@two:DD
z*:three@three:ZZ
zx@y:four@four:ZZZ
p*:five@five:PP
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<nobody@nobody>
RCPT TO:<one@one>
RCPT TO:<two@two>
RCPT TO:<three@three>
RCPT TO:<four@four>
RCPT TO:<five@five>
EOF

rm -f $tmp/rules
}
run_compare_test tests/rules-recip  <<END_OF_TEST_RESULTS
250 Sender='nobody@nobody'.^M
250 KK^M
553 DD^M
451 ZZ^M
250 Recipient='four@four'.^M
250 Recipient='five@five'.^M
END_OF_TEST_RESULTS


##### Test tests/rules-both #####

runtest() {
cat >$tmp/rules <<EOF
ka@a:a@b:K1
ka@a:b@b:K2
kb@a:a@b:K3
kb@a:b@b:K4
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<a@a>
RCPT TO:<a@b>
RCPT TO:<b@b>
MAIL FROM:<b@a>
RCPT TO:<a@b>
RCPT TO:<b@b>
EOF

rm -f $tmp/rules
}
run_compare_test tests/rules-both  <<END_OF_TEST_RESULTS
250 Sender='a@a'.^M
250 K1^M
250 K2^M
250 Sender='b@a'.^M
250 K3^M
250 K4^M
END_OF_TEST_RESULTS


##### Test tests/rules-multiline #####

runtest() {
cat >$tmp/rules <<EOF
ka@a:*:ONE\nTWO
ka@b:*:ONE\:TWO
ka@c:*:ONE\\\\TWO:
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<a@a>
MAIL FROM:<a@b>
MAIL FROM:<a@c>
EOF

rm -f $tmp/rules
}
run_compare_test tests/rules-multiline  <<END_OF_TEST_RESULTS
250-ONE^M
250 TWO^M
250 ONE:TWO^M
250 ONE\TWO^M
END_OF_TEST_RESULTS


##### Test tests/rules-list #####

runtest() {
cat >$tmp/rules <<EOF
k[[$tmp/list]]:*:LIST
k[[@$tmp/atlist]]:*:ATLIST
d*:*:DD
EOF

cat >$tmp/list <<EOF
a@a
EOF

cat >$tmp/atlist <<EOF
b
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<a@a>
MAIL FROM:<b@a>
MAIL FROM:<a@b>
MAIL FROM:<b@B>
EOF

rm -f $tmp/rules $tmp/list $tmp/atlist
}
run_compare_test tests/rules-list  <<END_OF_TEST_RESULTS
250 LIST^M
553 DD^M
250 ATLIST^M
250 ATLIST^M
END_OF_TEST_RESULTS


##### Test tests/rules-cdb #####

runtest() {
cat >$tmp/rules <<EOF
k[[$tmp/list.cdb]]:*:LIST
k[[@$tmp/atlist.cdb]]:*:ATLIST
d*:*:DD
EOF

cat <<EOF | cdbmake $tmp/list.cdb $tmp/list.tmp
+3,0:a@a->

EOF

cat <<EOF | cdbmake $tmp/atlist.cdb $tmp/atlist.tmp
+1,0:b->

EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<a@a>
MAIL FROM:<b@a>
MAIL FROM:<a@b>
MAIL FROM:<b@B>
EOF

rm -f $tmp/rules $tmp/list.cdb $tmp/atlist.cdb
}
run_compare_test tests/rules-cdb  <<END_OF_TEST_RESULTS
250 LIST^M
553 DD^M
250 ATLIST^M
250 ATLIST^M
END_OF_TEST_RESULTS


##### Test tests/rules-asterisk #####

runtest() {
cat >$tmp/rules <<EOF
ka@a:*:K1
k*@a:*:K2
kc@*:*:K3
k*:*:K4
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<a@a>
MAIL FROM:<b@a>
MAIL FROM:<c@b>
MAIL FROM:<d@d>
MAIL FROM:<>
EOF

rm -f $tmp/rules
}
run_compare_test tests/rules-asterisk  <<END_OF_TEST_RESULTS
250 K1^M
250 K2^M
250 K3^M
250 K4^M
250 K4^M
END_OF_TEST_RESULTS


##### Test tests/rules-empty #####

runtest() {
cat >$tmp/rules <<EOF
k:*:K1
k*:*:K2
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<>
MAIL FROM:<foo>
EOF

rm -f $tmp/rules
}
run_compare_test tests/rules-empty  <<END_OF_TEST_RESULTS
250 K1^M
250 K2^M
END_OF_TEST_RESULTS


##### Test tests/rules-defaultmsg #####

runtest() {
cat >$tmp/rules <<EOF
dd:*
zz:*
kk:*
d*:d
z*:z
k*:k
EOF

MAILRULES=$tmp/rules sfecho <<EOF
MAIL FROM:<d>
MAIL FROM:<z>
MAIL FROM:<k>
RCPT TO:<d>
RCPT TO:<z>
RCPT TO:<k>
EOF

rm -f $tmp/rules
}
run_compare_test tests/rules-defaultmsg  <<END_OF_TEST_RESULTS
553 Rejected^M
451 Deferred^M
250 OK^M
553 Rejected^M
451 Deferred^M
250 OK^M
END_OF_TEST_RESULTS


##### Test tests/smtpfront-content #####

runtest() {
RELAYCLIENT= sfecho <<EOF
MAIL FROM:<user@remote>
RCPT TO:<user@local>
DATA
Subject: test

foo
..
bar
.
MAIL FROM:<user@remote>
RCPT TO:<user@local>
DATA
Subject: test

foo
..
bar

.
EOF
}
run_compare_test tests/smtpfront-content  <<END_OF_TEST_RESULTS
250 Sender='user@remote'.^M
250 Recipient='user@local'.^M
354 End your message with a period.^M
250 Received 119 bytes.^M
250 Sender='user@remote'.^M
250 Recipient='user@local'.^M
354 End your message with a period.^M
250 Received 120 bytes.^M
END_OF_TEST_RESULTS


##### Test tests/smtpfront-databytes #####

runtest() {
DATABYTES=123 sfecho <<EOF
EHLO hostname
MAIL FROM:<somebody>
MAIL FROM:<somebody> SIZE
MAIL FROM:<somebody> SIZE=
MAIL FROM:<somebody> SIZE=100
MAIL FROM:<somebody> SIZE=123
MAIL FROM:<somebody> SIZE=124
RCPT TO:<nobody@nowhere>
EOF
}
run_compare_test tests/smtpfront-databytes  <<END_OF_TEST_RESULTS
250-unknown^M
250-SIZE 123^M
250-8BITMIME^M
250 PIPELINING^M
250 Sender='somebody'.^M
250 Sender='somebody'.^M
250 Sender='somebody'.^M
250 Sender='somebody'.^M
250 Sender='somebody'.^M
552 The message would exceed the maximum message size.^M
503 You must send MAIL FROM: first^M
END_OF_TEST_RESULTS


rm -rf $tmp
echo $tests_count tests executed, $tests_failed failures
if [ $tests_failed != 0 ]; then exit 1; fi
